/***************************************************************************/
/***************************************************************************/
/*                                                                         */
/*                    OSLab Module2: Memory Management                     */
/*                                                                         */
/***************************************************************************/
/***************************************************************************/

#include <dos.h>

#include "..\os.h"

#include "module2.h"

#define END_OF_LIST     -1
#define IN_USE          -2

/* Macro to determine control index from segment */

#define CONTROL_INDEX(x)        (((x >> 12) & 0xf) - 5)

/* Macro to determine offset from segment `x' */

#define OFFSET(x)     ((x & 0xfff) << 4)

/* Maximum number of segment groups */

#define MAX_SEGS      5

/* Size of each memory block */

#define S_512   512
#define S_1K    1024
#define S_2K    2048
#define S_4K    4096
#define S_16K   16384

/* Base address (index) of each group of memory blocks */

#define B_512   0
#define B_1K    128
#define B_2K    192
#define B_4K    224
#define B_16K   240

/* Number of each group of memory blocks */

#define N_512   128
#define N_1K    64
#define N_2K    32
#define N_4K    16
#define N_16K   4

/* Segment starting address of each group */

#define SEG_512         0x5000
#define SEG_1K          0x6000
#define SEG_2K          0x7000
#define SEG_4K          0x8000
#define SEG_16K         0x9000

PRIVATE int memory[N_512 + N_1K + N_2K + N_4K + N_16K];

struct control_entry
{
int seg_addr;   /* Segment of this segment group */
int free_list;  /* Index to first free segment of this size */
int limit;      /* Maximum number of segments in this group */
int base;       /* Index to first segment of this group in memory[] */
int size;       /* Size of this segment */
};

PRIVATE struct control_entry control[MAX_SEGS];

void initialize_segment(int seg, int seg_size, int base_index, int num_segs)
{
/*
 - initialize control[] and memory[] for a specific segment
 - parameters:
   * seg - segment base address (e.g. 0x5000)
   * seg_size - size of a segment (e.g., 512 or 1024)
   * base_index - starting index into memory[]
   * num_segs - number of segments of this particular size
 - control[] fields for this segment group are initialized
 - all memory[] elements are linked together (memory[i] points to i+1)
 - last element in list is assigned END_OF_LIST
*/
int index;
int i;

index = CONTROL_INDEX(seg);

control[index] . seg_addr = seg;
control[index] . free_list = base_index;
control[index] . limit = num_segs;
control[index] . base = base_index;
control[index] . size = seg_size;

for (i = 0; i < num_segs; i++)
        memory[base_index + i] = base_index + i + 1;

memory[base_index + num_segs - 1] = END_OF_LIST;

}

void memmgr_init()
{
/*
 - initialize memory[] and control[]
 - memory[] is an array of 5 separate linked lists, one per segment group
*/

initialize_segment(SEG_512, S_512, B_512, N_512);
initialize_segment(SEG_1K,  S_1K,  B_1K,  N_1K);
initialize_segment(SEG_2K,  S_2K,  B_2K,  N_2K); 
initialize_segment(SEG_4K,  S_4K,  B_4K,  N_4K); 
initialize_segment(SEG_16K, S_16K, B_16K, N_16K);

}

void *allocate(unsigned size)
{
/*
 - return a pointer to a block of memory at least size bytes long
 - two step process:
   1. scan control[] for first size that is available and fits
   2. if a segment exists, return MK_FP() of seg_addr and offset (from 
          free_list);
      if a segments doesn't exist, try next control[] and repeat the previous
          step.
 - if nothing available, return NULL
*/
unsigned offset;
int mem_index;
int segno;

/* Scan each segment group for size and availability */

for (segno = 0; segno < MAX_SEGS; segno++)
{
    if (size <= control[segno] . size && control[segno] . free_list != END_OF_LIST)
    {
        mem_index = control[segno] . free_list;
        control[segno] . free_list = memory[mem_index];
        memory[mem_index] = IN_USE;
        offset = (mem_index - control[segno] . base) * control[segno] . size;
        return MK_FP(control[segno] . seg_addr + (offset >> 4), 0);
    }
}

return NULL;

}

int deallocate(void *ptr)
{
/*
 - return ptr to memory if:
   1. the segment & 0xf000 is between 0x5000 and 0x9000
   2. the segment & 0x0fff:
      a) does not contain a fraction of a segment (e.g., a value between 512 
         and 1024);
      b) does not exceed the upper limit of segments for this group
      c) is currently IN_USE
      any of the above cause the deallocate to fail (return FALSE)
   3. otherwise, memory[] for the returned segment is reassigned to the 
      free list for control[].
*/
unsigned segment;
unsigned offset;
unsigned segno;
unsigned index;

segment = FP_SEG(ptr);

segno = CONTROL_INDEX(segment);

if (segno < MAX_SEGS)
{
        /* segno is valid (0 .. MAX_SEGS) -- now verify offset */
        offset = OFFSET(segment);

        /* is the offset valid?  no fractions of a segment */
        if (offset % control[segno] . size != 0)
                return FALSE;

        /* offset is valid; is it within the range of indexes? */
        if (offset / control[segno] . size >= control[segno] . limit)
                return FALSE;

        index = control[segno] . base + offset / control[segno] . size;

        /* is the specified index currently in use? */
        if (memory[index] != IN_USE)
                return FALSE;

        /* memory[index] is in use and can be returned */
        memory[index] = control[segno] . free_list;
        control[segno] . free_list = index;

        return TRUE;
}

/* Pointer contains an invalid segment */

return FALSE;

}
